import unittest
from defog.llm.utils import LLMResponse, LLM_COSTS_PER_TOKEN


class TestLLMResponse(unittest.TestCase):
    def test_token_costs(self):
        # Test data
        test_cases = [
            {
                "model_name": "gpt-4o",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["gpt-4o"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["gpt-4o"]["cached_input_cost_per1k"] * 0.5
                    + LLM_COSTS_PER_TOKEN["gpt-4o"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "gpt-4o-mini",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["gpt-4o-mini"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["gpt-4o-mini"]["cached_input_cost_per1k"]
                    * 0.5
                    + LLM_COSTS_PER_TOKEN["gpt-4o-mini"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "o1",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["o1"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["o1"]["cached_input_cost_per1k"] * 0.5
                    + LLM_COSTS_PER_TOKEN["o1"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "o1-mini",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["o1-mini"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["o1-mini"]["cached_input_cost_per1k"] * 0.5
                    + LLM_COSTS_PER_TOKEN["o1-mini"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "o3-mini",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["o3-mini"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["o3-mini"]["cached_input_cost_per1k"] * 0.5
                    + LLM_COSTS_PER_TOKEN["o3-mini"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "deepseek-chat",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["deepseek-chat"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["deepseek-chat"]["cached_input_cost_per1k"]
                    * 0.5
                    + LLM_COSTS_PER_TOKEN["deepseek-chat"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "deepseek-reasoner",
                "input_tokens": 1000,
                "cached_input_tokens": 500,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["deepseek-reasoner"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["deepseek-reasoner"][
                        "cached_input_cost_per1k"
                    ]
                    * 0.5
                    + LLM_COSTS_PER_TOKEN["deepseek-reasoner"]["output_cost_per1k"]
                ),
            },
            {
                "model_name": "claude-3-5-sonnet",
                "input_tokens": 1000,
                "cached_input_tokens": 0,
                "expected_cost": (
                    LLM_COSTS_PER_TOKEN["claude-3-5-sonnet"]["input_cost_per1k"]
                    + LLM_COSTS_PER_TOKEN["claude-3-5-sonnet"]["output_cost_per1k"]
                ),
            },
        ]

        for case in test_cases:
            with self.subTest(case=case):
                response = LLMResponse(
                    model=case["model_name"],
                    time=0.0,
                    input_tokens=case["input_tokens"],
                    cached_input_tokens=case["cached_input_tokens"],
                    output_tokens=1000,
                    content="",
                )
                response.__post_init__()  # Ensure cost calculation
                expected_cost_in_cents = case["expected_cost"] * 100  # Convert to cents
                self.assertAlmostEqual(
                    response.cost_in_cents, expected_cost_in_cents, places=10
                )


if __name__ == "__main__":
    unittest.main()



from defog.util import async_identify_categorical_columns, make_async_post_request
import asyncio
from io import StringIO
import pandas as pd
import json
from typing import List, Dict, Union


async def generate_postgres_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
    schemas: List[str] = ["public"],
) -> Union[Dict, List, str]:
    """
    Returns the schema of the tables in the database. Keys: column_name, data_type, column_description, custom_type_labels
    If tables is non-empty, we only generate the schema for the mentioned tables in the list.
    If schemas is non-empty, we only generate the schema for the mentioned schemas in the list.
    If return_tables_only is True, we return only the table names as a list.
    If upload is True, we send the schema to the defog servers and generate a CSV.
    If upload is False, we return the schema as a dict.
    If scan is True, we also scan the tables for categorical columns to enhance the column description.
    """
    try:
        import asyncpg
    except ImportError:
        raise ImportError(
            "asyncpg not installed. Please install it with `pip install psycopg2-binary`."
        )

    conn = await asyncpg.connect(**self.db_creds)
    schemas = tuple(schemas)

    if len(tables) == 0:
        # get all tables
        for schema in schemas:
            query = """
                SELECT table_name 
                FROM information_schema.tables 
                WHERE table_schema = $1;
            """
            rows = await conn.fetch(query, schema)
            if schema == "public":
                tables += [row[0] for row in rows]
            else:
                tables += [schema + "." + row[0] for row in rows]

    if return_tables_only:
        await conn.close()
        return tables

    print("Getting schema for each table that you selected...")

    table_columns = {}

    # get the columns for each table
    for schema in schemas:
        for table_name in tables:
            if "." in table_name:
                _, table_name = table_name.split(".", 1)
            query = """
                SELECT 
                    CAST(column_name AS TEXT), 
                    CAST(
                        CASE 
                            WHEN data_type = 'USER-DEFINED' THEN udt_name
                            ELSE data_type 
                        END AS TEXT
                    ) AS data_type,
                    col_description(
                        (quote_ident($2) || '.' || quote_ident($1))::regclass::oid, 
                        ordinal_position
                    ) AS column_description,
                    CASE
                        WHEN data_type = 'USER-DEFINED' THEN (
                            SELECT string_agg(enumlabel, ', ')
                            FROM pg_enum
                            WHERE enumtypid = (
                                SELECT oid
                                FROM pg_type
                                WHERE typname = udt_name
                            )
                        )
                        ELSE NULL
                    END AS custom_type_labels
                FROM information_schema.columns
                WHERE table_name = $1 AND table_schema = $2;
            """
            print(f"Schema for {schema}.{table_name}")
            rows = await conn.fetch(query, table_name, schema)
            rows = [row for row in rows]
            rows = [
                {
                    "column_name": row[0],
                    "data_type": row[1],
                    "column_description": row[2] or "",
                    "custom_type_labels": row[3].split(", ") if row[3] else [],
                }
                for row in rows
            ]
            if len(rows) > 0:
                if scan:
                    rows = await async_identify_categorical_columns(
                        conn=conn, cur=None, table_name=table_name, rows=rows
                    )
                if schema == "public":
                    table_columns[table_name] = rows
                else:
                    table_columns[schema + "." + table_name] = rows
    await conn.close()

    print(
        "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
    )
    if upload:
        # send the schemas dict to the defog servers
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": table_columns,
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return table_columns


async def generate_redshift_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    # when upload is True, we send the schema to the defog servers and generate a CSV
    # when its false, we return the schema as a dict
    try:
        import asyncpg
    except ImportError:
        raise ImportError(
            "asyncpg not installed. Please install it with `pip install psycopg2-binary`."
        )

    if "schema" not in self.db_creds:
        schema = "public"
        conn = await asyncpg.connect(**self.db_creds)
    else:
        schema = self.db_creds["schema"]
        del self.db_creds["schema"]
        conn = await asyncpg.connect(**self.db_creds)

    schemas = {}

    if len(tables) == 0:
        table_names_query = (
            "SELECT table_name FROM information_schema.tables WHERE table_schema = $1;"
        )
        results = await conn.fetch(table_names_query, schema)
        tables = [row[0] for row in results]

    if return_tables_only:
        await conn.close()
        return tables

    print("Getting schema for each table that you selected...")
    # get the schema for each table
    for table_name in tables:
        table_schema_query = "SELECT CAST(column_name AS TEXT), CAST(data_type AS TEXT) FROM information_schema.columns WHERE table_name::text = $1 AND table_schema= $2;"
        rows = await conn.fetch(table_schema_query, table_name, schema)
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if len(rows) > 0:
            if scan:
                await conn.execute(f"SET search_path TO {schema}")
                rows = await async_identify_categorical_columns(
                    conn=conn, cur=None, table_name=table_name, rows=rows
                )

            schemas[table_name] = rows

    await conn.close()

    if upload:
        print(
            "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        # send the schemas dict to the defog servers
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_mysql_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        import aiomysql
    except:
        raise Exception("aiomysql not installed.")

    conn = await aiomysql.connect(**self.db_creds)
    cur = await conn.cursor()
    schemas = {}

    if len(tables) == 0:
        # get all tables
        db_name = self.db_creds.get("database", "")
        await cur.execute(
            f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{db_name}';"
        )
        tables = [row[0] for row in await cur.fetchall()]

    if return_tables_only:
        await conn.close()
        return tables

    print("Getting schema for the relevant table in your database...")
    # get the schema for each table
    for table_name in tables:
        await cur.execute(
            "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = %s;",
            (table_name,),
        )
        rows = await cur.fetchall()
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if scan:
            rows = await async_identify_categorical_columns(
                conn=None,
                cur=cur,
                table_name=table_name,
                rows=rows,
                is_cursor_async=True,
            )
        if len(rows) > 0:
            schemas[table_name] = rows

    await conn.ensure_closed()

    if upload:
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_databricks_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        from databricks import sql
    except:
        raise Exception("databricks-sql-connector not installed.")

    conn = await asyncio.to_thread(sql.connect, **self.db_creds)
    schemas = {}
    async with await asyncio.to_thread(conn.cursor) as cur:
        print("Getting schema for each table that you selected...")
        # get the schema for each table

        if len(tables) == 0:
            # get all tables from databricks
            await asyncio.to_thread(
                cur.tables, schema_name=self.db_creds.get("schema", "default")
            )
            tables = [row.TABLE_NAME for row in await asyncio.to_thread(cur.fetchall())]

        if return_tables_only:
            await asyncio.to_thread(conn.close)
            return tables

        for table_name in tables:
            await asyncio.to_thread(
                cur.columns,
                schema_name=self.db_creds.get("schema", "default"),
                table_name=table_name,
            )
            rows = await asyncio.to_thread(cur.fetchall)
            rows = [row for row in rows]
            rows = [
                {"column_name": i.COLUMN_NAME, "data_type": i.TYPE_NAME} for i in rows
            ]
            if scan:
                rows = await async_identify_categorical_columns(
                    conn=None,
                    cur=cur,
                    table_name=table_name,
                    rows=rows,
                    is_cursor_async=False,
                )
            if len(rows) > 0:
                schemas[table_name] = rows

    await asyncio.to_thread(conn.close)

    if upload:
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_snowflake_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        import snowflake.connector
    except:
        raise Exception("snowflake-connector not installed.")

    conn = snowflake.connector.connect(
        user=self.db_creds["user"],
        password=self.db_creds["password"],
        account=self.db_creds["account"],
    )

    conn.cursor().execute(
        f"USE WAREHOUSE {self.db_creds['warehouse']}"
    )  # set the warehouse

    schemas = {}
    alt_types = {"DATE": "TIMESTAMP", "TEXT": "VARCHAR", "FIXED": "NUMERIC"}
    print("Getting schema for each table that you selected...")
    # get the schema for each table
    if len(tables) == 0:
        cur = conn.cursor()
        # get all tables from Snowflake database
        cur.execute_async("SHOW TERSE TABLES;")  # execute asynchrnously
        query_id = cur.sfqid  # get the query id to check the status
        while conn.is_still_running(conn.get_query_status(query_id)):
            await asyncio.sleep(1)
        res = cur.fetchall()
        tables = [f"{row[3]}.{row[4]}.{row[1]}" for row in res]

    if return_tables_only:
        conn.close()
        return tables

    for table_name in tables:
        rows = []
        cur = conn.cursor()
        cur.execute_async(f"SHOW COLUMNS IN {table_name};")
        query_id = cur.sfqid
        while conn.is_still_running(conn.get_query_status(query_id)):
            await asyncio.sleep(1)
        rows = cur.fetchall()
        rows = [
            {
                "column_name": i[2],
                "data_type": json.loads(i[3])["type"],
                "column_description": i[8],
            }
            for i in rows
        ]
        for idx, row in enumerate(rows):
            if row["data_type"] in alt_types:
                row["data_type"] = alt_types[row["data_type"]]
            rows[idx] = row
        cur = conn.cursor()
        if scan:
            rows = await async_identify_categorical_columns(
                conn=conn,
                cur=cur,
                table_name=table_name,
                rows=rows,
                is_cursor_async=False,
                db_type="snowflake",
            )
        cur.close()
        if len(rows) > 0:
            schemas[table_name] = rows

    conn.close()

    if upload:
        print(
            "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_bigquery_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        from google.cloud import bigquery
    except:
        raise Exception("google-cloud-bigquery not installed.")

    client = await asyncio.to_thread(
        bigquery.Client.from_service_account_json, self.db_creds["json_key_path"]
    )
    project_id = [p.project_id for p in await asyncio.to_thread(client.list_projects)][
        0
    ]
    datasets = [
        dataset.dataset_id for dataset in await asyncio.to_thread(client.list_datasets)
    ]
    schemas = {}

    if len(tables) == 0:
        # get all tables
        tables = []
        for dataset in datasets:
            table_list = await asyncio.to_thread(client.list_tables, dataset)
            tables += [
                f"{project_id}.{dataset}.{table.table_id}" for table in table_list
            ]

    print("Getting the schema for each table that you selected...")
    # get the schema for each table
    for table_name in tables:
        table = await asyncio.to_thread(client.get_table, table_name)
        rows = table.schema
        rows = [{"column_name": i.name, "data_type": i.field_type} for i in rows]
        if len(rows) > 0:
            schemas[table_name] = rows

    await asyncio.to_thread(client.close)

    if upload:
        print(
            "Sending the schema to Defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_sqlserver_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    return_tables_only: bool = False,
) -> str:
    try:
        import aioodbc
    except Exception as e:
        raise Exception(
            "aioodbc not installed. Please install it with `pip install aioodbc`."
        )
    if self.db_creds["database"] != "":
        connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};DATABASE={self.db_creds['database']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"
    else:
        connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"

    conn = await aioodbc.connect(dsn=connection_string)
    cur = await conn.cursor()
    schemas = {}
    schema = self.db_creds.get("schema", "dbo")

    if len(tables) == 0:
        # get all tables
        await cur.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = %s;",
            (schema,),
        )
        if schema == "dbo":
            tables = [row[0] for row in await cur.fetchall()]
        else:
            tables = [schema + "." + row[0] for row in await cur.fetchall()]

    if return_tables_only:
        await conn.close()
        return tables

    print("Getting schema for the relevant table in your database...")
    # get the schema for each table
    for table_name in tables:
        await cur.execute(
            "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = %s;",
            (table_name,),
        )
        rows = await cur.fetchall()
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if len(rows) > 0:
            schemas[table_name] = rows
    await conn.close()
    if upload:
        print(
            "Sending the schema to Defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        resp = await make_async_post_request(
            url=f"{self.base_url}/get_schema_csv",
            payload={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
        )
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


async def generate_db_schema(
    self,
    tables: list,
    scan: bool = True,
    upload: bool = True,
    return_tables_only: bool = False,
    return_format: str = "csv",
) -> str:
    if self.db_type == "postgres":
        return await self.generate_postgres_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "mysql":
        return await self.generate_mysql_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "bigquery":
        return await self.generate_bigquery_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "redshift":
        return await self.generate_redshift_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "snowflake":
        return await self.generate_snowflake_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "databricks":
        return await self.generate_databricks_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "sqlserver":
        return await self.generate_sqlserver_schema(
            tables,
            return_format=return_format,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    else:
        raise ValueError(
            f"Creation of a DB schema for {self.db_type} is not yet supported via the library. If you are a premium user, please contact us at founder@defog.ai so we can manually add it."
        )



import unittest
import pytest
from defog.llm.utils import chat_async, chat_openai, chat_anthropic, chat_gemini
from defog.llm.utils_function_calling import get_function_specs
from pydantic import BaseModel, Field
import aiohttp
from io import StringIO
import json

# ==================================================================================================
# Functions for function calling
# ==================================================================================================

IO_STREAM = StringIO()


def log_to_file(function_name, input_args, tool_result):
    """
    Simple function to test logging to a StringIO object.
    Used in test_post_tool_calls_openai and test_post_tool_calls_anthropic
    """
    sorted_input_args = {k: input_args[k] for k in sorted(input_args)}

    message = {
        "function_name": function_name,
        "args": sorted_input_args,
        "result": tool_result,
    }
    message = json.dumps(message, indent=4)
    IO_STREAM.write(message + "\n")
    return IO_STREAM.getvalue()


class WeatherInput(BaseModel):
    latitude: float = Field(default=0.0, description="The latitude of the location")
    longitude: float = Field(default=0.0, description="The longitude of the location")


async def get_weather(input: WeatherInput):
    """
    This function returns the current temperature (in celsius) for the given latitude and longitude.
    """
    async with aiohttp.ClientSession() as client:
        r = await client.get(
            f"https://api.open-meteo.com/v1/forecast?latitude={input.latitude}&longitude={input.longitude}&current=temperature_2m,wind_speed_10m&hourly=temperature_2m,relative_humidity_2m,wind_speed_10m",
        )
        return_object = await r.json()
        return return_object["current"]["temperature_2m"]


class Numbers(BaseModel):
    a: int = 0
    b: int = 0


def numsum(input: Numbers):
    """
    This function returns the sum of two numbers
    """
    return input.a + input.b


def numprod(input: Numbers):
    """
    This function returns the product of two numbers
    """
    return input.a * input.b


# ==================================================================================================
# Tests
# ==================================================================================================
class TestGetFunctionSpecs(unittest.TestCase):
    def setUp(self):
        self.openai_model = "gpt-4o"
        self.anthropic_model = "claude-3-5-sonnet-20241022"
        self.tools = [get_weather, numsum, numprod]
        self.maxDiff = None
        self.openai_specs = [
            {
                "type": "function",
                "function": {
                    "name": "get_weather",
                    "description": "This function returns the current temperature (in celsius) for the given latitude and longitude.",
                    "parameters": {
                        "properties": {
                            "latitude": {
                                "description": "The latitude of the location",
                                "type": "number",
                            },
                            "longitude": {
                                "description": "The longitude of the location",
                                "type": "number",
                            },
                        },
                        "type": "object",
                    },
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "numsum",
                    "description": "This function returns the sum of two numbers",
                    "parameters": {
                        "properties": {
                            "a": {"type": "integer"},
                            "b": {"type": "integer"},
                        },
                        "type": "object",
                    },
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "numprod",
                    "description": "This function returns the product of two numbers",
                    "parameters": {
                        "properties": {
                            "a": {"type": "integer"},
                            "b": {"type": "integer"},
                        },
                        "type": "object",
                    },
                },
            },
        ]
        self.anthropic_specs = [
            {
                "name": "get_weather",
                "description": "This function returns the current temperature (in celsius) for the given latitude and longitude.",
                "input_schema": {
                    "properties": {
                        "latitude": {
                            "description": "The latitude of the location",
                            "type": "number",
                        },
                        "longitude": {
                            "description": "The longitude of the location",
                            "type": "number",
                        },
                    },
                    "type": "object",
                },
            },
            {
                "name": "numsum",
                "description": "This function returns the sum of two numbers",
                "input_schema": {
                    "properties": {
                        "a": {"type": "integer"},
                        "b": {"type": "integer"},
                    },
                    "type": "object",
                },
            },
            {
                "name": "numprod",
                "description": "This function returns the product of two numbers",
                "input_schema": {
                    "properties": {
                        "a": {"type": "integer"},
                        "b": {"type": "integer"},
                    },
                    "type": "object",
                },
            },
        ]

    def test_get_function_specs(self):
        openai_specs = get_function_specs(self.tools, self.openai_model)
        anthropic_specs = get_function_specs(self.tools, self.anthropic_model)

        self.assertEqual(openai_specs, self.openai_specs)
        self.assertEqual(anthropic_specs, self.anthropic_specs)


class TestToolUseFeatures(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.tools = [get_weather, numsum, numprod]
        self.weather_qn = "What is the current temperature in Singapore? Return the answer as a number and nothing else."
        self.weather_qn_specific = "What is the current temperature in Singapore? Singapore's latitude is 1.3521 and longitude is 103.8198. Return the answer as a number and nothing else."
        self.arithmetic_qn = "What is the product of 31283 and 2323, added to 5? Always use the tools provided for all calculation, even simple calculations. Return only the final answer, nothing else."
        self.arithmetic_answer = "72670414"
        self.arithmetic_expected_tool_outputs = [
            {"name": "numprod", "args": {"a": 31283, "b": 2323}, "result": 72670409},
            {"name": "numsum", "args": {"a": 72670409, "b": 5}, "result": 72670414},
        ]

    @pytest.mark.asyncio
    async def test_tool_use_arithmetic_async_openai(self):
        result = await chat_async(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
        )
        print(result)
        self.assertEqual(result.content, self.arithmetic_answer)
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})

    @pytest.mark.asyncio
    async def test_tool_use_weather_async_openai(self):
        result = await chat_async(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn,
                },
            ],
            tools=self.tools,
            max_retries=1,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    @pytest.mark.asyncio
    async def test_tool_use_arithmetic_async_anthropic(self):
        result = await chat_async(
            model="claude-3-5-sonnet-20241022",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        self.assertEqual(result.content, self.arithmetic_answer)

    @pytest.mark.asyncio
    async def test_tool_use_weather_async_anthropic(self):
        result = await chat_async(
            model="claude-3-5-sonnet-20241022",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn,
                },
            ],
            tools=self.tools,
            max_retries=1,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    @pytest.mark.asyncio
    async def test_tool_use_arithmetic_async_gemini(self):
        result = await chat_async(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
        )
        print(result)
        self.assertEqual(result.content, self.arithmetic_answer)
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})

    @pytest.mark.asyncio
    async def test_tool_use_weather_async_gemini(self):
        result = await chat_async(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn_specific,
                },
            ],
            tools=self.tools,
            max_retries=1,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertEqual(
            result.tool_outputs[0]["args"], {"latitude": 1.3521, "longitude": 103.8198}
        )
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    @pytest.mark.asyncio
    async def test_post_tool_calls_openai(self):
        result = await chat_async(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
            post_tool_function=log_to_file,
        )
        print(result)
        self.assertEqual(result.content, self.arithmetic_answer)
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})
        expected_stream_value = (
            json.dumps(
                {
                    "function_name": "numprod",
                    "args": {"a": 31283, "b": 2323},
                    "result": 72670409,
                },
                indent=4,
            )
            + "\n"
            + json.dumps(
                {
                    "function_name": "numsum",
                    "args": {"a": 72670409, "b": 5},
                    "result": 72670414,
                },
                indent=4,
            )
            + "\n"
        )
        self.assertEqual(IO_STREAM.getvalue(), expected_stream_value)

        # clear IO_STREAM
        IO_STREAM.seek(0)
        IO_STREAM.truncate()

    async def test_post_tool_calls_anthropic(self):
        result = await chat_async(
            model="claude-3-5-sonnet-latest",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
            post_tool_function=log_to_file,
        )
        print(result)
        self.assertEqual(result.content, self.arithmetic_answer)
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})
        expected_stream_value = (
            json.dumps(
                {
                    "function_name": "numprod",
                    "args": {"a": 31283, "b": 2323},
                    "result": 72670409,
                },
                indent=4,
            )
            + "\n"
            + json.dumps(
                {
                    "function_name": "numsum",
                    "args": {"a": 72670409, "b": 5},
                    "result": 72670414,
                },
                indent=4,
            )
            + "\n"
        )
        self.assertEqual(IO_STREAM.getvalue(), expected_stream_value)

        # clear IO_STREAM
        IO_STREAM.seek(0)
        IO_STREAM.truncate()

    @pytest.mark.asyncio
    async def test_post_tool_calls_gemini(self):
        result = await chat_async(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": self.arithmetic_qn,
                },
            ],
            tools=self.tools,
            post_tool_function=log_to_file,
        )
        print(result)
        self.assertEqual(result.content, self.arithmetic_answer)
        for expected, actual in zip(
            self.arithmetic_expected_tool_outputs, result.tool_outputs
        ):
            self.assertEqual(expected["name"], actual["name"])
            self.assertEqual(expected["args"], actual["args"])
            self.assertEqual(expected["result"], actual["result"])
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum", "numprod"})
        expected_stream_value = (
            json.dumps(
                {
                    "function_name": "numprod",
                    "args": {"a": 31283, "b": 2323},
                    "result": 72670409,
                },
                indent=4,
            )
            + "\n"
            + json.dumps(
                {
                    "function_name": "numsum",
                    "args": {"a": 72670409, "b": 5},
                    "result": 72670414,
                },
                indent=4,
            )
            + "\n"
        )
        self.assertEqual(IO_STREAM.getvalue(), expected_stream_value)

        # clear IO_STREAM
        IO_STREAM.seek(0)
        IO_STREAM.truncate()

    def test_async_tool_in_sync_function_openai(self):
        result = chat_openai(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn,
                },
            ],
            tools=self.tools,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    def test_async_tool_in_sync_function_anthropic(self):
        result = chat_anthropic(
            model="claude-3-5-sonnet-20241022",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn,
                },
            ],
            tools=self.tools,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    def test_async_tool_in_sync_function_gemini(self):
        result = chat_gemini(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": self.weather_qn_specific,
                },
            ],
            tools=self.tools,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"get_weather"})
        self.assertEqual(result.tool_outputs[0]["name"], "get_weather")
        self.assertEqual(
            result.tool_outputs[0]["args"], {"latitude": 1.3521, "longitude": 103.8198}
        )
        self.assertGreaterEqual(float(result.content), 21)
        self.assertLessEqual(float(result.content), 38)

    def test_required_tool_choice_openai(self):
        result = chat_openai(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": "Give me the result of 102 + 2",
                },
            ],
            tools=self.tools,
            tool_choice="required",
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum"})
        self.assertEqual(result.tool_outputs[0]["name"], "numsum")
        self.assertIn("104", result.content.lower())

    def test_required_tool_choice_anthropic(self):
        result = chat_anthropic(
            model="claude-3-5-sonnet-20241022",
            messages=[
                {
                    "role": "user",
                    "content": "Give me the result of 102 + 2",
                },
            ],
            tools=self.tools,
            tool_choice="required",
            max_completion_tokens=1000,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum"})
        self.assertEqual(result.tool_outputs[0]["name"], "numsum")
        self.assertIn("104", result.content.lower())

    def test_required_tool_choice_gemini(self):
        result = chat_gemini(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": "If I add 1 egg to 1 egg, how many eggs are there in total?",
                },
            ],
            tools=self.tools,
            tool_choice="required",
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertSetEqual(set(tools_used), {"numsum"})
        self.assertEqual(result.tool_outputs[0]["name"], "numsum")
        self.assertIn("2", result.content.lower())

    def test_forced_tool_choice_openai(self):
        """
        This test forces the use of numprod even though the user question asks for addition.
        """
        result = chat_openai(
            model="gpt-4o",
            messages=[
                {
                    "role": "user",
                    "content": "Give me the result of 102 + 2",
                },
            ],
            tools=self.tools,
            tool_choice="numprod",
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertIn("numprod", set(tools_used))
        self.assertEqual(result.tool_outputs[0]["name"], "numprod")
        self.assertEqual(result.tool_outputs[0]["result"], 204)

    def test_forced_tool_choice_anthropic(self):
        """
        This test forces the use of numprod even though the user question asks for addition.
        """
        result = chat_anthropic(
            model="claude-3-5-sonnet-20241022",
            messages=[
                {
                    "role": "user",
                    "content": "Give me the result of 102 + 2",
                },
            ],
            tools=self.tools,
            tool_choice="numprod",
            max_completion_tokens=1000,
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertIn("numprod", set(tools_used))
        self.assertEqual(result.tool_outputs[0]["name"], "numprod")
        self.assertEqual(result.tool_outputs[0]["result"], 204)

    def test_forced_tool_choice_gemini(self):
        """
        This test forces the use of numprod even though the user question asks for addition.
        """
        result = chat_gemini(
            model="gemini-2.0-flash",
            messages=[
                {
                    "role": "user",
                    "content": "Give me the result of 102 + 2",
                },
            ],
            tools=self.tools,
            tool_choice="numprod",
        )
        print(result)
        tools_used = [output["name"] for output in result.tool_outputs]
        self.assertIn("numprod", set(tools_used))
        self.assertEqual(result.tool_outputs[0]["name"], "numprod")
        self.assertEqual(result.tool_outputs[0]["result"], 204)

    def test_invalid_forced_tool_choice_openai(self):
        """
        This test forces the use of an invalid tool `sum` and checks that an error is raised
        """
        with self.assertRaises(ValueError) as context:
            result = chat_openai(
                model="gpt-4o",
                messages=[
                    {
                        "role": "user",
                        "content": self.arithmetic_qn,
                    },
                ],
                tools=self.tools,
                tool_choice="sum",
            )
        self.assertEqual(
            str(context.exception),
            "Forced function `sum` is not in the list of provided tools",
        )

    def test_invalid_forced_tool_choice_anthropic(self):
        """
        This test forces the use of an invalid tool `sum` and checks that an error is raised
        """
        with self.assertRaises(ValueError) as context:
            result = chat_anthropic(
                model="claude-3-5-sonnet-20241022",
                messages=[
                    {
                        "role": "user",
                        "content": self.arithmetic_qn,
                    },
                ],
                tools=self.tools,
                tool_choice="sum",
                max_completion_tokens=1000,
            )
        self.assertEqual(
            str(context.exception),
            "Forced function `sum` is not in the list of provided tools",
        )

    def test_invalid_forced_tool_choice_gemini(self):
        """
        This test forces the use of an invalid tool `sum` and checks that an error is raised
        """
        with self.assertRaises(ValueError) as context:
            result = chat_gemini(
                model="gemini-2.0-flash",
                messages=[
                    {
                        "role": "user",
                        "content": self.arithmetic_qn,
                    },
                ],
                tools=self.tools,
                tool_choice="sum",
            )
        self.assertEqual(
            str(context.exception),
            "Forced function `sum` is not in the list of provided tools",
        )



import os
from chalice import Chalice

app = Chalice(app_name="defog-lambda")

from defog import Defog
import json

creds64_str = os.environ["DEFOG_CREDS_64"]
defog = Defog(
    base64creds=creds64_str, save_json=False
)  # don't save the json file as no file system in lambda


@app.route("/", methods=["POST"])
def answer():
    # This is the JSON body the user sent in their POST request.
    query = app.current_request.json_body
    print(f"query = {query}")
    question = query["question"]
    hard_filters = query.get("hard_filters")
    answer = defog.run_query(question, hard_filters=hard_filters)
    answer = json.dumps(answer, default=str)
    return answer
